;; Generated package description from spell-fu.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "spell-fu" "0.3.0.20260521.52528" "Fast & light spelling highlighter" '((emacs "29.1")) :commit "e1bb525d726eab0146f112a9c80ec79d1e6be3d9" :authors '(("Campbell Barton" . "ideasman42@gmail.com")) :maintainer '("Campbell Barton" . "ideasman42@gmail.com") :keywords '("convenience") :url "https://codeberg.org/ideasman42/emacs-spell-fu")
